#!/usr/bin/python3
"""
## © University of Southampton IT Innovation Centre, 2018
##
## Copyright in this software belongs to University of Southampton
## IT Innovation Centre of Gamma House, Enterprise Road,
## Chilworth Science Park, Southampton, SO16 7NS, UK.
##
## This software may not be used, sold, licensed, transferred, copied
## or reproduced in whole or in part in any manner or form or in or
## on any media by any person other than in accordance with the terms
## of the Licence Agreement supplied with the software, or otherwise
## without the prior written consent of the copyright owners.
##
## This software is distributed WITHOUT ANY WARRANTY, without even the
## implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
## PURPOSE, except where stated in the Licence Agreement supplied with
## the software.
##
##      Created By :            Nikolay Stanchev
##      Created Date :          04-06-2018
##      Created for Project :   FLAME
"""

from threading import Event
from unittest import mock
from clmcservice.aggregation.aggregator import AggregatorThread
from clmcservice.aggregation.influx_data_interface import MockResultSet, network_result_item, service_result_item, drop_timestamp, generate_e2e_no_timestamp_row


class TestAggregation(object):
    """
    A unit test to ensure the functionality of the aggregator is correct.
    """

    ACTUAL_RESULTS = "actual_aggregated_results"  # the attribute name of the actual results data structure
    EXPECTED_RESULTS = "expected_aggregated_results"  # the attribute name of the expected results data structure
    FINISHED = "finished_event"  # the attribute name of the flag object, which marks the end of the test

    def points_generator(self, network_items, service_items):
        """
        A generator method intended to be used by the mock db client when involving the mocked query() method. It takes the network and service items, and generates a result from
        those items each time query() is called by taking turns - starts with network result, followed by service result and then it repeats, until all items have been exhausted.
        Network items and service items are expected to have the same length.

        :param network_items: the network data to generate from
        :param service_items: the service data to generate from

        :return: a generator object
        """

        assert len(network_items) == len(service_items), "The data points generator must receive the same number of network items as the number of service items"
        index = 0

        while not getattr(self, self.FINISHED).is_set():
            items = network_items[index]
            yield MockResultSet(items)

            items = service_items[index]

            # before yielding the service data points, check if both sets of data points are enumerated
            if index == len(network_items)-1:
                # if so, set the finished flag of the test
                getattr(self, self.FINISHED).set()

            yield MockResultSet(items)

            index += 1

    def setup_mock_db_client(self, mock_class):
        """
        Sets up a mock db client and also defines the expected aggregation results from the test.

        :param mock_class: the mock class used as an influx db client instance
        :return:
        """

        setattr(self, self.ACTUAL_RESULTS, [])  # initially, there are no actual results, these are built progressively while the aggregator is running
        setattr(self, self.EXPECTED_RESULTS, [
            generate_e2e_no_timestamp_row(path_id="SR1-SR3", source_sfr="SR1", target_sfr="SR3", endpoint="endpoint1", sf_instance="ms1.flame.org", delay_forward=10,
                                          delay_reverse=15, delay_service=10, avg_request_size=1024, avg_response_size=8, avg_bandwidth=1200),
            generate_e2e_no_timestamp_row(path_id="SR1-SR3", source_sfr="SR1", target_sfr="SR3", endpoint="endpoint1", sf_instance="ms1.flame.org", delay_forward=5,
                                          delay_reverse=25, delay_service=40, avg_request_size=16, avg_response_size=2048, avg_bandwidth=1200),
            generate_e2e_no_timestamp_row(path_id="SR1-SR2", source_sfr="SR1", target_sfr="SR2", endpoint="endpoint2", sf_instance="ms2.flame.org", delay_forward=15,
                                          delay_reverse=35, delay_service=60, avg_request_size=32, avg_response_size=1024, avg_bandwidth=1200),
            generate_e2e_no_timestamp_row(path_id="SR4-SR5", source_sfr="SR4", target_sfr="SR5", endpoint="endpoint5", sf_instance="ms5.flame.org", delay_forward=11,
                                          delay_reverse=25, delay_service=50, avg_request_size=2048, avg_response_size=32, avg_bandwidth=1200),
            generate_e2e_no_timestamp_row(path_id="SR1-SR2", source_sfr="SR1", target_sfr="SR2", endpoint="endpoint2", sf_instance="ms2.flame.org", delay_forward=12,
                                          delay_reverse=5, delay_service=60, avg_request_size=32, avg_response_size=1024, avg_bandwidth=1200),
            generate_e2e_no_timestamp_row(path_id="SR1-SR3", source_sfr="SR1", target_sfr="SR3", endpoint="endpoint1", sf_instance="ms1.flame.org", delay_forward=16,
                                          delay_reverse=25, delay_service=40, avg_request_size=16, avg_response_size=2048, avg_bandwidth=1200),
            generate_e2e_no_timestamp_row(path_id="SR10-SR12", source_sfr="SR12", target_sfr="SR10", endpoint="endpoint10", sf_instance="ms4.flame.org", delay_forward=22,
                                          delay_reverse=3, delay_service=75, avg_request_size=1024, avg_response_size=64, avg_bandwidth=1200),
            generate_e2e_no_timestamp_row(path_id="SR14-SR15", source_sfr="SR14", target_sfr="SR15", endpoint="endpoint15", sf_instance="ms2.flame.org", delay_forward=24,
                                          delay_reverse=27, delay_service=105, avg_request_size=1024, avg_response_size=128, avg_bandwidth=1200),
            generate_e2e_no_timestamp_row(path_id="SR14-SR15", source_sfr="SR15", target_sfr="SR14", endpoint="endpoint14", sf_instance="ms1.flame.org", delay_forward=27,
                                          delay_reverse=24, delay_service=85, avg_request_size=32, avg_response_size=64, avg_bandwidth=1200),
        ])  # defines the expected rows from the aggregation
        setattr(self, self.FINISHED, Event())

        # initialises the influx data generator, which is involved each time the query() method of the mock db client is called
        mock_points = self.points_generator(
            # network items is a list of tuples, each tuple represents a result from a query; each time query() is called and a network measurement must be generated, then one of
            # these tuples is generated, empty tuple means result with no points
            network_items=[
                (
                    network_result_item("network_delays", "SR1-SR3", "SR1", "SR3", 10, 1200),
                    network_result_item("network_delays", "SR1-SR3", "SR3", "SR1", 15, 1200),
                    network_result_item("network_delays", "SR1-SR33", "SR1", "SR33", 15, 1200),
                    network_result_item("network_delays", "SR2-SR11", "SR11", "SR2", 15, 1200)
                ),
                (
                    network_result_item("network_delays", "SR1-SR3", "SR1", "SR3", 5, 1200),
                    network_result_item("network_delays", "SR1-SR3", "SR3", "SR1", 25, 1200),
                    network_result_item("network_delays", "SR1-SR2", "SR1", "SR2", 15, 1200),
                    network_result_item("network_delays", "SR1-SR2", "SR2", "SR1", 35, 1200)
                ),
                (
                    network_result_item("network_delays", "SR4-SR5", "SR4", "SR5", 5, 1200),
                    network_result_item("network_delays", "SR4-SR5", "SR5", "SR4", 25, 1200),
                ),
                (),
                (
                    network_result_item("network_delays", "SR4-SR5", "SR4", "SR5", 5, 1200),
                    network_result_item("network_delays", "SR4-SR5", "SR5", "SR4", 25, 1200),
                    network_result_item("network_delays", "SR0-SR1", "SR0", "SR1", 5, 1200),
                    network_result_item("network_delays", "SR0-SR1", "SR1", "SR0", 25, 1200),
                    network_result_item("network_delays", "SR10-SR12", "SR10", "SR12", 11, 1200),
                    network_result_item("network_delays", "SR10-SR12", "SR12", "SR10", 22, 1200),
                    network_result_item("network_delays", "SR14-SR15", "SR14", "SR15", 24, 1200),
                    network_result_item("network_delays", "SR14-SR15", "SR15", "SR14", 26, 1200),
                ),
                (
                    network_result_item("network_delays", "SR4-SR5", "SR4", "SR5", 11, 1200),
                    network_result_item("network_delays", "SR1-SR2", "SR1", "SR2", 12, 1200),
                    network_result_item("network_delays", "SR1-SR2", "SR2", "SR1", 5, 1200),
                    network_result_item("network_delays", "SR1-SR3", "SR1", "SR3", 16, 1200),
                    network_result_item("network_delays", "SR10-SR12", "SR10", "SR12", 3, 1200),
                    network_result_item("network_delays", "SR14-SR15", "SR15", "SR14", 27, 1200),
                )
            ],
            # service items is a list of tuples, each tuple represents a result from a query; each time query() is called and a service measurement must be generated, then one of
            # these tuples is generated, empty tuple means result with no points
            service_items=[
                (
                    service_result_item("service_delays", "SR3", "endpoint1", "ms1.flame.org", 10, 1024, 8),
                    service_result_item("service_delays", "SR33", "endpoint33", "ms2.flame.org", 20, 4096, 8),
                    service_result_item("service_delays", "SR11", "endpoint11", "ms3.flame.org", 30, 1024, 8),
                ),
                (
                    service_result_item("service_delays", "SR3", "endpoint1", "ms1.flame.org", 40, 16, 2048),
                    service_result_item("service_delays", "SR2", "endpoint2", "ms2.flame.org", 60, 32, 1024)
                ),
                (
                    service_result_item("service_delays", "SR6", "endpoint6", "ms1.flame.org", 60, 1024, 8),
                    service_result_item("service_delays", "SR7", "endpoint7", "ms1.flame.org", 70, 1024, 8),
                ),
                (
                    service_result_item("service_delays", "SR6", "endpoint6", "ms1.flame.org", 65, 2048, 16),
                    service_result_item("service_delays", "SR8", "endpoint8", "ms2.flame.org", 75, 2048, 16),
                    service_result_item("service_delays", "SR9", "endpoint9", "ms3.flame.org", 25, 2048, 16),
                ),
                (),
                (
                    service_result_item("service_delays", "SR5", "endpoint5", "ms5.flame.org", 50, 2048, 32),
                    service_result_item("service_delays", "SR10", "endpoint10", "ms4.flame.org", 75, 1024, 64),
                    service_result_item("service_delays", "SR15", "endpoint15", "ms2.flame.org", 105, 1024, 128),
                    service_result_item("service_delays", "SR14", "endpoint14", "ms1.flame.org", 85, 32, 64),
                )
            ]
        )

        # implement the query() and write_points() methods of the mock db client
        mock_class.query = lambda query: next(mock_points)  # query() returns the next element of the mock_points generator
        mock_class.write_points = lambda points: getattr(self, self.ACTUAL_RESULTS).append(drop_timestamp(points[0]))  # write_points() adds aggregated rows to actual results list

        # in the end of the test, we can compare the expected results with the actual results that were generated during the aggregation process

    @mock.patch('clmcservice.aggregation.aggregator.InfluxDBClient', autospec=True)
    def test_aggregator(self, MockDBClient):
        """
        The actual test that's executed when running pytest.

        :param MockDBClient: a mock object argument passed by the mock.patch decorator. The decorator changes all occurrences of InfluxDBClient in the aggregator's code to the
        return value of this MockDBClient object
        """

        # set up the mock db client by providing implementations for the necessary methods (query and write_points)
        self.setup_mock_db_client(MockDBClient.return_value)

        # start the aggregator as a thread, report period set to 2 so that the unit tests is not taking too long
        t = AggregatorThread(report_period=2)
        t.start()

        # wait until the finished flag has been set
        getattr(self, self.FINISHED).wait()

        # stop the thread when the aggregation has finished
        t.stop()

        # compare the expected results with teh actual results that were collected during the aggregation process
        expected_results = getattr(self, self.EXPECTED_RESULTS)
        actual_results = getattr(self, self.ACTUAL_RESULTS)
        assert type(actual_results) is list
        assert type(expected_results) is list
        assert len(actual_results) == len(expected_results), "Actual and expected result differ in length."

        # we compare sorted versions of the expected and actual results; this is because the aggregator implementation uses dictionary for efficiency purposes, hence the order of
        # the collected results may vary, especially on different OS; hence, we only care about the two list of results to contain the same elements
        assert sorted(actual_results, key=lambda k: k['tags']['path_ID']) == sorted(expected_results, key=lambda k: k['tags']['path_ID']), \
            "Test failure - aggregation process returns incorrect results."
